/*
 * Program to receive packets for Multicast groups.
 * 
 *
 * Written by Polly Huang, University of Southern California, July 1995.
 * Bug fixes and modifications by Pavlin Radoslavov, May 1997
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <net/if.h>
#include <sys/ioctl.h>
#include <netinet/in.h>
#include "mtest.h"

main( argc, argv )
    int argc;
    char **argv;
{
    int so;
    char line[MAX_LINESIZE];
    char *lineptr;
    struct ip_mreq imr;			/* the multicast join/leave address */
    unsigned i1, i2, i3, i4, g1, g2, g3, g4;
    unsigned e1, e2, e3, e4, e5, e6;	/* the ethernet address */
    struct in_addr dest_group;		/* in_addr + port# = sockaddr_in */
    struct sockaddr_in dest;
    u_short port_no;
    int addlen;           /* address length for bind() */
    Packet packet;
    int n;                /* return value of system call */
    long received_bytes;
    int on = 1;		  /* A flag */
    
    bzero(&imr, sizeof(imr));
    bzero(&dest_group, sizeof(dest_group));
    bzero(&dest, sizeof(dest));
    bzero(&packet, sizeof(packet));
    
    if( (so = socket( AF_INET, SOCK_DGRAM, 0 )) < 0) {
	perror("can't open socket");
	exit (1);
    }
    dest.sin_family = AF_INET;
    
    printf("multicast receiving test program; ");
    printf("enter ? for list of commands\n");
    
    while (fgets(line, MAX_LINESIZE-1, stdin) != NULL) {
	lineptr = line;
	while (*lineptr == ' ' || *lineptr == '\t')
	    lineptr++;
	
	switch (*lineptr ) {
	case '?':
	    printf( "%s%s%s%s",
		    " j g.g.g.g i.i.i.i              - join  IP  multicast group \n",
		    " l g.g.g.g i.i.i.i              - leave IP  multicast group \n",
		    " r g.g.g.g port#                - receive from multicast interface \n",
		    " q                              - quit                      \n\n" );
	    break;
	case 'r':
	    lineptr++;
	    while (*lineptr == ' ' || *lineptr == '\t')
		lineptr++;
	    if ( (n = sscanf(lineptr, "%u.%u.%u.%u %hd",
			     &g1, &g2, &g3, &g4, &port_no)) != 5) {
		printf("bad args\n");
		break;
	    }
	    dest_group.s_addr = (g1<<24) | (g2<<16) | (g3<<8) | g4;
	    dest_group.s_addr = htonl(dest_group.s_addr);
#if 1
	    /* TODO: Doesn't work for Solaris 5.4 */
	    dest.sin_addr = dest_group;
#endif
	    dest.sin_port = htons(port_no);
	    addlen = sizeof(struct sockaddr_in);
	    if (setsockopt(so, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on)) < 0)
		perror("Error in setsockopt(SO_REUSEADDR)");
#ifdef SO_REUSEPORT
	    if (setsockopt(so, SOL_SOCKET, SO_REUSEPORT, &on, sizeof(on)) < 0)
		perror("Error in setsockopt(SO_REUSEPORT)");
#endif	    
	    if (bind(so, (struct sockaddr *)&dest, addlen) < 0)
		perror("Error in binding socket");
	    else
		printf ("Bind socket\n");
	    do {
		if( (received_bytes =
		     recvfrom(so, (char *)&packet, sizeof(packet), 0, 
			      NULL, &addlen)) < 0) {
		    perror("Error in receiving pkt");
		} else {
		    if (received_bytes == ntohs(packet.length)) {
			struct timeval time_total;
			struct timeval *time_start;
			struct timeval *time_end;
			
			time_start = (struct timeval *)((char *)&packet + 12);
			time_end   = time_start +1;
			time_start->tv_sec = ntohl(time_start->tv_sec);
			time_start->tv_usec = ntohl(time_start->tv_usec);
			time_end->tv_sec = ntohl(time_end->tv_sec);
			time_end->tv_usec = ntohl(time_end->tv_usec);
			time_total.tv_sec = 0;
			time_total.tv_usec = 0;
			MY_DELTA_TIME(*time_end, *time_start, time_total);
			printf ("Receive pkt # %d, for mgroup %u.%u.%u.%u port %hd, length=%d, t_sec=%ld, t_usec=%ld\n", ntohs(packet.seq_num), packet.mdestadd[0], packet.mdestadd[1], packet.mdestadd[2], packet.mdestadd[3], ntohs(packet.port_no), ntohs(packet.length), time_total.tv_sec, time_total.tv_usec);
		    }
		    else
			printf ("Receive pkt # %d, for mgroup %u.%u.%u.%u port %hd, length=%d, but MUST BE %d\n", ntohs(packet.seq_num), packet.mdestadd[0], packet.mdestadd[1], packet.mdestadd[2], packet.mdestadd[3], ntohs(packet.port_no), received_bytes, ntohs(packet.length));
		}
	    } while (1);
	    break;
	    
	case 'j':
	    lineptr++;
	    while (*lineptr == ' ' || *lineptr == '\t' )
		lineptr++;
	    if ((n = sscanf( lineptr, "%u.%u.%u.%u %u.%u.%u.%u %u",
			     &g1, &g2, &g3, &g4, &i1, &i2, &i3, &i4 )) != 8 ) {
		printf("bad args\n");
		break;
	    }
	    imr.imr_multiaddr.s_addr = (g1<<24) | (g2<<16) | (g3<<8) | g4;
	    imr.imr_multiaddr.s_addr = htonl(imr.imr_multiaddr.s_addr);
	    imr.imr_interface.s_addr = (i1<<24) | (i2<<16) | (i3<<8) | i4;
	    imr.imr_interface.s_addr = htonl(imr.imr_interface.s_addr);
	    if (setsockopt(so, IPPROTO_IP, IP_ADD_MEMBERSHIP,
			   (char *)&imr, sizeof(struct ip_mreq) ) < 0)
		perror( "can't join group" );
	    else
		printf( "group joined\n" );
	    break;
	    
	case 'l':
	    lineptr++;
	    while (*lineptr == ' ' || *lineptr == '\t')
		lineptr++;
	    if ((n = sscanf(lineptr, "%u.%u.%u.%u %u.%u.%u.%u %u",
			    &g1, &g2, &g3, &g4, &i1, &i2, &i3, &i4 )) != 8) {
		printf("bad args\n");
		break;
	    }
	    imr.imr_multiaddr.s_addr = (g1<<24) | (g2<<16) | (g3<<8) | g4;
	    imr.imr_multiaddr.s_addr = htonl(imr.imr_multiaddr.s_addr);
	    imr.imr_interface.s_addr = (i1<<24) | (i2<<16) | (i3<<8) | i4;
	    imr.imr_interface.s_addr = htonl(imr.imr_interface.s_addr);
	    if (setsockopt(so, IPPROTO_IP, IP_DROP_MEMBERSHIP,
			   (char *)&imr, sizeof(struct ip_mreq) ) < 0)
		perror("can't leave group");
	    else
		printf("group left\n");
	    break;
	    
	case 'q':
	    exit (0);
	    
	case '\n':
	    break;
	    
	default:
	    printf("bad command\n");
	    break;
	}
    }
}
