/*
 * Program to send packets to Multicast groups.
 * 
 *
 * Written by Puneet Sharma and Polly Huang, 
 * University of Southern California, June 1995.
 * Bug fixes and small modifications by Pavlin Radoslavov, May 1997
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <net/if.h>
#include <sys/ioctl.h>
#include <netinet/in.h>
#include "mtest.h"


main( argc, argv )
    int argc;
    char **argv;
{
    int so;
    char line[MAX_LINESIZE];
    char *lineptr, *charptr;
    int i, n, tmp, tmp2;
    unsigned i1, i2, i3, i4, g1, g2, g3, g4;
    unsigned e1, e2, e3, e4, e5, e6;		/* ethernet address */
    struct in_addr dest_group;			/* destination group addr */
    struct sockaddr_in dest;			/* in_addr + port# */
    u_short port_no;
    struct in_addr interface_addr;		/* interface addr */
    u_char ttl;					/* time_to_live */
    unsigned size_min = 996, size_max = 996;	/* default packet size */
    int setint = 0;				/* default set pkt interval */
    float interval_min = 1.0, interval_max = 1.0; /* default packet interval */
    u_long interval = 1000000;			/* default interval in microsec */
    float rate = 1.0;				/* default packet rate */
    int setpktnum = 1;				/* default duration in packets */
    u_long duration = INFINITE;			/* default duration packets */
    int num_pkts = 10;				/* default duration packets */
    struct timeval time;			/* random number gen. seed */
    int inx;	/* seq number in sending loop (set pkt as duration case */
    u_long time_passed;	/* time passed in sending loop (set time as duration case */
    Packet packet;				/* packet to send */
    
    
    bzero(&dest_group, sizeof(dest_group));
    bzero(&dest, sizeof(dest));
    bzero(&interface_addr, sizeof(interface_addr));
    bzero(&time, sizeof(time));
    bzero(&packet, sizeof(packet));
    
    for (i = (unsigned)&(packet.data[0]) - (unsigned)&packet,
	     charptr=packet.data;
	 i < sizeof(packet.data); i++, charptr++) {
	*charptr = i&0xff;
    }
    
    gettimeofday(&time, NULL);
    srand(time.tv_sec*1000000 + time.tv_usec);
    
    if ((so = socket( AF_INET, SOCK_DGRAM, 0 )) < 0) {
	perror("can't open socket");
	exit (1);
    }
    dest.sin_family = AF_INET;
    
    printf("multicast sending test program; ");
    printf("enter ? for list of commands\n");
    while (fgets(line, MAX_LINESIZE-1, stdin ) != NULL) {
	lineptr = line;
	while (*lineptr == ' ' || *lineptr == '\t')
	    lineptr++;
	switch (*lineptr) {
	case '?':
	    printf( "%s%s%s%s%s%s%s%s%s%s%s",
		    " i i.i.i.i                      - set the multicast interface \n",
		    " t ttl_value                    - set the ttl_value for multicast \n",
		    " z min-max                      - set packet size range in byte \n",
		    " v min-max                      - set inter-packet interval range in second\n",
		    "                                  ex. v 0.5-2 or v 1-1 for fixed interval\n",
		    " a rate                         - set packet sending rate\n",
		    " d type #pkts_or_#sec           - set communication duration\n",
		    "                                  type=1 set #pkt ex. d 1 20\n",
		    "                                  type=2 set #sec ex. d 0 25\n",
		    " s g.g.g.g port#                - send to IP  multicast group \n",
		    " q                              - quit                      \n\n" );
	    break;
	    
	case 'i': 
	    lineptr++;
	    while (*lineptr == ' ' || *lineptr == '\t')
		lineptr++;
	    if ((n = sscanf(lineptr, "%u.%u.%u.%u %u",
			    &i1, &i2, &i3, &i4)) != 4) {
		printf("bad args\n");
		break;
	    }
	    interface_addr.s_addr = (i1<<24) | (i2<<16) | (i3<<8) | i4;
	    interface_addr.s_addr = htonl(interface_addr.s_addr);
	    if (setsockopt(so, IPPROTO_IP, IP_MULTICAST_IF, 
			   (char *)&interface_addr,
			   sizeof(struct in_addr) ) < 0) {
		perror("can't set the sending interface");
	    }
	    else
		printf( "interface set\n" );
	    break;
	    
	case 't':
	    lineptr++;
	    while (*lineptr == ' ' || *lineptr == '\t')
		lineptr++;
	    if ((n = sscanf( lineptr, "%u", &tmp)) != 1) {
		printf( "bad args\n" );
		break;
	    }
	    ttl = (u_char) tmp;
	    if (setsockopt(so, IPPROTO_IP, IP_MULTICAST_TTL, 
			   (char *) &ttl, sizeof(ttl) ) < 0) {
		perror( "can't set the ttl" );
	    }
	    else
		printf("ttl set to %d \n", ttl);
	    break;
	    
	case 'z':
	    ++lineptr;
	    while (*lineptr == ' ' || *lineptr == '\t')
		lineptr++;
	    if ((n = sscanf( lineptr, "%u-%u",
			     &size_min, &size_max)) != 2) {
		printf("bad args\n");
		break;
	    }
	    else
		printf("packet size range set: %u-%u bytes\n",
		       size_min, size_max);
	    break;
	    
	case 'v':
	    lineptr++;
	    while (*lineptr == ' ' || *lineptr == '\t')
		lineptr++;
	    if ((n = sscanf(lineptr, "%f-%f",
			    &interval_min, &interval_max)) != 2) {
		printf("bad args\n");
		break;
	    }
	    else {
		setint = 1;
		printf("packet interval range set: %.3f - %.3f seconds\n",
		       interval_min, interval_max);
	    }
	    break;
	    
	case 'a':
	    lineptr++;
	    while (*lineptr == ' ' || *lineptr == '\t' )
		lineptr++;
	    if ( (n = sscanf(lineptr, "%f ", &rate)) != 1) {
		printf( "bad args\n" );
		break;
	    }
	    else {
		setint = 0;
		printf("sending rate set: %.3f KBps\n", rate);
	    }
	    break;
	    
	case 'd':
	    lineptr++;
	    while (*lineptr == ' ' || *lineptr == '\t')
		lineptr++;
	    if ( (n = sscanf(lineptr, "%d %d ", &setpktnum, &tmp2)) != 2) {
		printf("bad args\n");
		break;
	    }
	    else {
		if (setpktnum == 1) {
		    num_pkts = tmp2;
		    duration = INFINITE;
		    printf ("packet number set: %d packets\n", num_pkts);
		}
		else {
		    num_pkts = 99999;
		    duration = (long)tmp2 * 1000000;
		    printf("time duration set: %d seconds\n", tmp2);
		}
	    }
	    break;
	    
 	case 's':
	    lineptr++;
	    while (*lineptr == ' ' || *lineptr == '\t')
		lineptr++;
	    if ( (n = sscanf(lineptr, "%u.%u.%u.%u %hd",
			     &g1, &g2, &g3, &g4, &port_no)) != 5) {
		printf( "bad args\n" );
		break;
	    }
	    dest_group.s_addr = (g1<<24) | (g2<<16) | (g3<<8) | g4;
	    dest_group.s_addr = htonl(dest_group.s_addr);
	    dest.sin_addr = dest_group;
	    dest.sin_port = htons(port_no);
	    
	    packet.mdestadd[0] = g1;
	    packet.mdestadd[1] = g2;
	    packet.mdestadd[2] = g3;
	    packet.mdestadd[3] = g4;
	    packet.port_no = htons(port_no);
	    if (setpktnum == 1)
		printf ("Sending... for %d packets\n", num_pkts);
	    else
		printf ("Sending... for %d seconds\n", (int)(duration/1000000));
	    
	    inx = 0;
	    time_passed = 0;
	    while (inx < num_pkts && time_passed < duration) {
		packet.seq_num = htons(inx);
		packet.length = htons((unsigned)&(packet.data[0]) - (unsigned)&packet + rand()%(size_max-size_min+1) + size_min);
		
		if (setint == 1)
		    interval =
			rand()%((long)((interval_max-interval_min)*1000000 +
				       1)) +(long)(interval_min*1000000);
		else
		    interval = (long)((ntohs(packet.length))/(rate *1024)*1000000);
		
		if (sendto(so, (char *)&packet, ntohs(packet.length), 0, 
			   (struct sockaddr *)&dest, sizeof(struct sockaddr_in)) < 0) {
		    perror("Error in sending pkt");
		} else {
		    printf("Sent pkt # %d, for mgroup %u.%u.%u.%u port %hd, length=%u, interval=%d microseconds\n", inx, packet.mdestadd[0], packet.mdestadd[1], packet.mdestadd[2], packet.mdestadd[3], ntohs(packet.port_no), ntohs(packet.length), interval);
		}
		
		usleep(interval);
		inx++;
		time_passed += interval;
	    }
	    
	    packet.seq_num = htons(9999);
	    packet.length = htons(0);
	    break;
	    
	case 'q':
	    exit (0);
	    
	case '\n':
	    break;
	    
	default:
	    printf("bad command\n");
	    break;
	}
    }
}
